"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ProtectRecordingDelegate = void 0;
const protect_ffmpeg_1 = require("./protect-ffmpeg");
const protect_timeshift_1 = require("./protect-timeshift");
// Camera recording delegate implementation for Protect.
class ProtectRecordingDelegate {
    constructor(protectCamera) {
        var _a;
        this.accessory = protectCamera.accessory;
        this.api = protectCamera.api;
        this.timeshift = new protect_timeshift_1.ProtectTimeshiftBuffer(protectCamera);
        this.debug = protectCamera.platform.debug.bind(protectCamera.platform);
        this.ffmpegStream = null;
        this.isInitialized = false;
        this._isRecording = false;
        this.log = protectCamera.platform.log;
        this.name = protectCamera.name.bind(protectCamera);
        this.nvr = protectCamera.nvr;
        this.protectCamera = protectCamera;
        this.maxRecordingDuration = parseInt((_a = this.nvr.optionGet(this.accessory.context.device, "Video.HKSV.Recording.MaxDuration")) !== null && _a !== void 0 ? _a : "0");
        this.recordedSegments = 0;
        this.rtspEntry = null;
        this.transmitListener = null;
    }
    // Process HomeKit requests to activate or deactivate HKSV recording capabilities for a camera.
    async updateRecordingActive(active) {
        // If we are no longer recording, stop the livestream.
        if (!active) {
            this.timeshift.stop();
            // Inform the user of the state change, if needed.
            if (this.isRecording !== active) {
                this.log.info("%s: Disabling HomeKit Secure Video event recording.", this.name());
            }
            this._isRecording = active;
            return;
        }
        // We have no recording configuration available yet. Set our desired state and we're done.
        // Once we have a recording configuration, we'll get called again and be able to begin timeshifting.
        if (!this.recordingConfig) {
            this._isRecording = active;
            return;
        }
        // Figure out which camera channel we should use for the livestream based on the requested resolution.
        this.rtspEntry = this.protectCamera.findRtsp(this.recordingConfig.videoCodec.resolution[0], this.recordingConfig.videoCodec.resolution[1], this.accessory.context.device);
        if (!this.rtspEntry) {
            this._isRecording = false;
            this.log.error("%s: Unable to start the HomeKit Secure Video timeshift buffer: no valid RTSP stream profile was found.", this.name());
            return;
        }
        // Set the bitrate to what HomeKit is looking for. This is particularly useful when we occasionally have
        // to livestream to a user, where bitrates can be different and even get reconfigured in realtime. By
        // contrast, HomeKit Secure Video has a consistent bitrate it accepts, and we want to try to match it as
        // closely as posible.
        if (!(await this.protectCamera.setBitrate(this.rtspEntry.channel.id, this.recordingConfig.videoCodec.parameters.bitRate * 1000))) {
            this.log.error("%s: Unable to set the bitrate to %skbps for HomeKit Secure Video event recording.", this.name(), this.recordingConfig.videoCodec.parameters.bitRate);
            return;
        }
        // Fire up the timeshift buffer.
        if (!(await this.timeshift.start(this.rtspEntry.channel.id))) {
            this.log.error("%s: Unable to start the timeshift buffer for HomeKit Secure Video.", this.name());
            return;
        }
        // Inform the user of the state change, if needed.
        if ((this._isRecording !== active) || !this.isInitialized) {
            this.isInitialized = true;
            this.log.info("%s: Enabling HomeKit Secure Video event recording: %sx%s@%sfps, %s kbps with a %s second timeshift buffer.", this.name(), this.rtspEntry.resolution[0], this.rtspEntry.resolution[1], this.rtspEntry.resolution[2], this.recordingConfig.videoCodec.parameters.bitRate, this.timeshift.length / 1000);
            // Inform the user if there's a maximum event recording duration set.
            if (this.maxRecordingDuration) {
                this.log.info("%s: HomeKit Secure Video recordings will be no longer than ~%s seconds.", this.name(), this.maxRecordingDuration);
            }
        }
        // Update our recording state internally.
        this._isRecording = active;
    }
    // Process updated recording configuration settings from HomeKit Secure Video.
    updateRecordingConfiguration(configuration) {
        // If we're set to an undefined state, it's because HomeKit can't figure out a valid configuration to use.
        // This is typically due to a factory reset of the camera or a similar edge case. We choose to handle it
        // by stopping our timeshift buffer.
        if (!configuration) {
            this.recordingConfig = configuration;
            this.timeshift.stop();
            return;
        }
        // Save the new recording configuration.
        this.recordingConfig = configuration;
        // Tell our timeshift buffer how many seconds HomeKit has requested we prebuffer.
        this.timeshift.length = this.recordingConfig.prebufferLength;
        // Start or restart our timeshift buffer based on our updated configuration.
        void this.updateRecordingActive(this.isRecording);
    }
    // Handle the actual recording stream request.
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    async *handleRecordingStreamRequest(streamId) {
        // If we've explicitly disabled HKSV recording, we're done right now.
        if (!this.accessory.context.hksvRecording) {
            const streamHeader = await this.timeshift.getStreamHeader();
            if (streamHeader) {
                yield { data: streamHeader, isLast: true };
                return;
            }
            // We should never get to this point, but in case we do, this is a failsafe to ensure that no video
            // recording can ever take place if we've disabled it through the HKSV recording switch accessory.
            // It will generate a warning in Homebridge that can be ignored, but it will ensure nothing can be recorded.
            yield { data: Buffer.alloc(0), isLast: true };
            return;
        }
        // Start transmitting our timeshift buffer and process it through FFmpeg.
        if (!this.startTransmitting() || !this.ffmpegStream) {
            // Something went wrong, so we gracefully tell HomeKit we have no packets to transmit and return.
            this.stopTransmitting();
            yield { data: Buffer.alloc(0), isLast: true };
            return;
        }
        for await (const segment of this.ffmpegStream.generator()) {
            let isLastSegment = false;
            // If we've stopped transmitting - make sure we don't send anything else.
            if (!this.timeshift.isTransmitting) {
                return;
            }
            // No segment doesn't mean we're done necessarily, but it does mean we need to wait for FFmpeg to catch up.
            if (!segment) {
                continue;
            }
            // If we've exceeded a user-configured maximum recording duration, let HomeKit know we're stopping.
            if (this.maxRecordingDuration && ((this.recordedSegments * this.timeshift.segmentLength) / 1000) > this.maxRecordingDuration) {
                isLastSegment = true;
            }
            yield {
                data: segment,
                isLast: isLastSegment
            };
            if (isLastSegment) {
                return;
            }
        }
    }
    // Receive an acknowledgement from HomeKit that it's seen an end-of-stream packet from us.
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    acknowledgeStream(streamId) {
        // Since HomeKit knows our transmission is ending, it's safe to do so now.
        this.stopTransmitting();
    }
    // Process HomeKit requests to end the transmission of the recording stream.
    closeRecordingStream(streamId, reason) {
        this.stopTransmitting(reason);
    }
    // Start transmitting to the HomeKit hub our timeshifted fMP4 stream.
    startTransmitting() {
        var _a;
        // If there's a prior instance of FFmpeg, clean up after ourselves.
        if (this.ffmpegStream) {
            this.ffmpegStream.stop();
            this.ffmpegStream = null;
        }
        // If there's a prior instance of our transmit handler, clean it up.
        if (this.transmitListener) {
            this.timeshift.removeListener("segment", this.transmitListener);
            this.transmitListener = null;
        }
        // If we don't have a recording configuration from HomeKit or an RTSP profile, we can't continue.
        if (!this.recordingConfig || !this.rtspEntry) {
            return false;
        }
        // Start a new FFmpeg instance to transcode using HomeKit's requirements.
        // We want to keep feeding HomeKit until it tells us it's finished, or we decide we don't want to send anymore
        // fMP4 packets. We treat this the same was a DVR works where you can pause live television, but it continues to
        // buffer what's being broadcast until you're ready to watch it. This is the same idea.
        // Keep track of how many fMP4 segments we are feeding FFmpeg.
        this.recordedSegments = 0;
        // Check to see if the user has audio enabled or disabled for recordings.
        const isAudioActive = ((_a = this.protectCamera.stream.controller.recordingManagement) === null || _a === void 0 ? void 0 : _a.recordingManagementService.getCharacteristic(this.api.hap.Characteristic.RecordingAudioActive).value) === 1 ? true : false;
        this.ffmpegStream = new protect_ffmpeg_1.FfmpegRecordingProcess(this.protectCamera, this.recordingConfig, isAudioActive);
        // Start the livestream.
        const timeshiftedSeconds = this.timeshift.length / 1000;
        // Let the timeshift buffer know it's time to transmit and continue timeshifting.
        this.timeshift.transmitStream(true);
        let seenFtyp = false;
        let seenMoov = false;
        // Listen in for events from the timeshift buffer and feed FFmpeg. This looks simple, conceptually,
        // but there's a lot going on here.
        this.transmitListener = (segment) => {
            var _a, _b;
            (_b = (_a = this.ffmpegStream) === null || _a === void 0 ? void 0 : _a.stdin) === null || _b === void 0 ? void 0 : _b.write(segment);
            // We don't want the fMP4 ftyp or moov boxes accounted for in our recording statistics so
            // let's filter them out.
            if (!seenFtyp && this.timeshift.isFtyp(segment)) {
                seenFtyp = true;
                return;
            }
            if (!seenMoov && this.timeshift.isMoov(segment)) {
                seenMoov = true;
                return;
            }
            this.recordedSegments++;
        };
        this.timeshift.on("segment", this.transmitListener);
        // Inform the user.
        this.log.debug("%s: Beginning a HomeKit Secure Video recording event with a timeshift buffer of %s seconds.", this.name(), timeshiftedSeconds);
        return true;
    }
    // Stop transmitting the HomeKit hub our timeshifted fMP4 stream.
    stopTransmitting(reason) {
        // We're done transmitting, so we can go back to maintaining our timeshift buffer for HomeKit.
        this.timeshift.transmitStream(false);
        // Kill any FFmpeg sessions.
        if (this.ffmpegStream) {
            this.ffmpegStream.stop();
            this.ffmpegStream = null;
        }
        if (this.transmitListener) {
            this.timeshift.removeListener("segment", this.transmitListener);
            this.transmitListener = null;
        }
        // Inform the user.
        if (this.accessory.context.hksvRecording) {
            this.log.info("%s: HomeKit Secure Video has recorded a %s second motion event.", this.name(), Math.round((this.recordedSegments * this.timeshift.segmentLength) / 1000));
        }
        // If we have a reason for stopping defined, and it's noteworthy, inform the user.
        let reasonDescription;
        switch (reason) {
            case 3 /* CANCELLED */:
                reasonDescription = "HomeKit canceled the request.";
                break;
            case 5 /* UNEXPECTED_FAILURE */:
                reasonDescription = "An unexpected protocol failure has occured.";
                break;
            case 6 /* TIMEOUT */:
                reasonDescription = "The request timed out.";
                break;
            default:
                break;
        }
        if ((reason !== undefined) && (reason !== 0 /* NORMAL */)) {
            this.log.error("%s: HomeKit Secure Video event recording ended abnormally: %s", this.name(), reasonDescription);
        }
    }
    // Return our HomeKit Secure Video recording state. This effectively tells us if HKSV has been configured and is on.
    get isRecording() {
        return this._isRecording;
    }
    // Return our current HomeKit Secure Video recording configuration.
    get recordingConfiguration() {
        var _a;
        return (_a = this.recordingConfig) !== null && _a !== void 0 ? _a : null;
    }
}
exports.ProtectRecordingDelegate = ProtectRecordingDelegate;
//# sourceMappingURL=protect-record.js.map